@@ -1,13 +1,18 @@
+import argparse
 import atexit
 import json
 import logging
 import os
+import shutil
 import ssl
 import subprocess
 import sys
+import tempfile
 import time
 import urllib
 from abc import ABCMeta, abstractmethod
+from enum import Enum
+from pathlib import Path
 from typing import Any, Dict, Optional, Tuple
 
 import aiohttp
@@ -18,19 +23,22 @@ from unifi.core import RetryableError
 AVClientRequest = AVClientResponse = Dict[str, Any]
 
 
-class UnifiCamBase(metaclass=ABCMeta):
-    def __init__(self, args, logger=None):
-        self.args = args
-        if logger is None:
-            self.logger = logging.getLogger(__class__)
-        else:
-            self.logger = logger
+class SmartDetectObjectType(Enum):
+    PERSON = "person"
+    CAR = "car"
 
-        self._msg_id = 0
-        self._init_time = time.time()
-        self._streams = {}
-        self._motion_event_id = 0
-        self._motion_event_active = False
+
+class UnifiCamBase(metaclass=ABCMeta):
+    def __init__(self, args: argparse.Namespace, logger: logging.Logger) -> None:
+        self.args = args
+        self.logger = logger
+
+        self._msg_id: int = 0
+        self._init_time: float = time.time()
+        self._streams: Dict[str, str] = {}
+        self._motion_snapshot = None
+        self._motion_event_id: int = 0
+        self._motion_event_ts: Optional[float] = None
         self._ffmpeg_handles = {}
 
         # Set up ssl context for requests
@@ -38,15 +46,15 @@ class UnifiCamBase(metaclass=ABCMeta):
         self._ssl_context.check_hostname = False
         self._ssl_context.verify_mode = ssl.CERT_NONE
         self._ssl_context.load_cert_chain(args.cert, args.cert)
-        self._session: Optional[websockets.client.WebSocketClientProtocol] = None
+        self._session: Optional[websockets.legacy.client.WebSocketClientProtocol] = None
         atexit.register(self.close_streams)
 
     @classmethod
-    def add_parser(cls, parser):
+    def add_parser(cls, parser: argparse.ArgumentParser) -> None:
         parser.add_argument(
             "--ffmpeg-args",
             "-f",
-            default="-f lavfi -i aevalsrc=0  -vcodec copy -strict -2 -c:a aac",
+            default="-c:v copy -ar 32000 -ac 2 -codec:a aac -b:a 32k",
             help="Transcoding args for `ffmpeg -i <src> <args> <dst>`",
         )
         parser.add_argument(
@@ -56,7 +64,7 @@ class UnifiCamBase(metaclass=ABCMeta):
             help="RTSP transport protocol used by stream",
         )
 
-    async def _run(self, ws):
+    async def _run(self, ws) -> None:
         self._session = ws
         await self.init_adoption()
         while True:
@@ -71,93 +79,127 @@ class UnifiCamBase(metaclass=ABCMeta):
                     self.logger.info("Reconnecting...")
                     raise RetryableError()
 
-    async def run(self):
+    async def run(self) -> None:
         pass
 
-    def get_video_settings(self):
+    def get_video_settings(self) -> Dict[str, Any]:
         return {}
 
-    def change_video_settings(self, options):
+    def change_video_settings(self, options) -> None:
         pass
 
     @abstractmethod
-    async def get_snapshot(self):
+    async def get_snapshot(self) -> Path:
         raise NotImplementedError("You need to write this!")
 
     @abstractmethod
-    def get_stream_source(self, stream_index: str):
+    def get_stream_source(self, stream_index: str) -> str:
         raise NotImplementedError("You need to write this!")
 
-    @abstractmethod
-    def start_video_stream(
-        self, stream_index: str, stream_name: str, destination: Tuple[str, int]
-    ):
-        raise NotImplementedError("You need to write this!")
+    def get_feature_flags(self) -> Dict[str, Any]:
+        return {
+            "mic": True,
+        }
 
-    @abstractmethod
-    def stop_video_stream(
-        self,
-        stream_index: str,
-    ):
-        raise NotImplementedError("You need to write this!")
+    # API for subclasses
+    async def trigger_motion_start(
+        self, object_type: Optional[SmartDetectObjectType] = None
+    ) -> None:
+        if not self._motion_event_ts:
+            payload: Dict[str, Any] = {
+                "clockBestMonotonic": 0,
+                "clockBestWall": 0,
+                "clockMonotonic": int(self.get_uptime()),
+                "clockStream": int(self.get_uptime()),
+                "clockStreamRate": 1000,
+                "clockWall": int(round(time.time() * 1000)),
+                "edgeType": "start",
+                "eventId": self._motion_event_id,
+                "eventType": "motion",
+                "levels": {"0": 47},
+                "motionHeatmap": "",
+                "motionSnapshot": "",
+            }
+            if object_type:
+                payload.update(
+                    {
+                        "objectTypes": [object_type.value],
+                        "edgeType": "enter",
+                        "zonesStatus": {"0": 48},
+                        "smartDetectSnapshot": "",
+                    }
+                )
 
-    ### API for subclasses
-    async def trigger_motion_start(self) -> None:
-        if not self._motion_event_active:
+            self.logger.info(f"Triggering motion start (idx: {self._motion_event_id})")
             await self.send(
                 self.gen_response(
-                    "EventAnalytics",
-                    payload={
-                        "clockBestMonotonic": 0,
-                        "clockBestWall": 0,
-                        "clockMonotonic": int(self.get_uptime()),
-                        "clockStream": int(self.get_uptime()),
-                        "clockStreamRate": 1000,
-                        "clockWall": int(round(time.time() * 1000)),
-                        "edgeType": "start",
-                        "eventId": self._motion_event_id,
-                        "eventType": "motion",
-                        "levels": {"0": 47},
-                        "motionHeatmap": "",
-                        "motionSnapshot": "",
-                    },
+                    "EventSmartDetect" if object_type else "EventAnalytics",
+                    payload=payload,
                 ),
             )
-            self._motion_event_active = True
+            self._motion_event_ts = time.time()
+            self._motion_snapshot = shutil.copyfile(
+                await self.get_snapshot(), tempfile.NamedTemporaryFile(delete=True).name
+            )
 
-    async def trigger_motion_stop(self) -> None:
-        if self._motion_event_active:
+    async def trigger_motion_stop(
+        self, object_type: Optional[SmartDetectObjectType] = None
+    ) -> None:
+        motion_start_ts = self._motion_event_ts
+        if motion_start_ts:
+            payload: Dict[str, Any] = {
+                "clockBestMonotonic": int(self.get_uptime()),
+                "clockBestWall": int(round(motion_start_ts * 1000)),
+                "clockMonotonic": int(self.get_uptime()),
+                "clockStream": int(self.get_uptime()),
+                "clockStreamRate": 1000,
+                "clockWall": int(round(time.time() * 1000)),
+                "edgeType": "stop",
+                "eventId": self._motion_event_id,
+                "eventType": "motion",
+                "levels": {"0": 49},
+                "motionHeatmap": "heatmap.png",
+                "motionSnapshot": "motionsnap.jpg",
+            }
+            if object_type:
+                payload.update(
+                    {
+                        "objectTypes": [object_type.value],
+                        "edgeType": "leave",
+                        "zonesStatus": {"0": 48},
+                        "smartDetectSnapshot": "motionsnap.jpg",
+                    }
+                )
+            self.logger.info(f"Triggering motion stop (idx: {self._motion_event_id})")
             await self.send(
                 self.gen_response(
-                    "EventAnalytics",
-                    payload={
-                        "clockBestMonotonic": int(self.get_uptime()),
-                        "clockBestWall": 1618540558190,
-                        "clockMonotonic": int(self.get_uptime()),
-                        "clockStream": int(self.get_uptime()),
-                        "clockStreamRate": 1000,
-                        "clockWall": int(round(time.time() * 1000)),
-                        "edgeType": "stop",
-                        "eventId": self._motion_event_id,
-                        "eventType": "motion",
-                        "levels": {"0": 49},
-                        "motionHeatmap": f"heatmap_00000{self._motion_event_id}.png",
-                        "motionSnapshot": f"motionsnap_00000{self._motion_event_id}.jpg",
-                    },
+                    "EventSmartDetect" if object_type else "EventAnalytics",
+                    payload=payload,
                 ),
             )
             self._motion_event_id += 1
-            self._motion_event_active = False
+            self._motion_event_ts = None
 
-    ### Protocol implementation
+    def update_motion_snapshot(self, path: Path) -> None:
+        self._motion_snapshot = path
 
+    async def fetch_to_file(self, url: str, dst: Path) -> bool:
+        try:
+            async with aiohttp.request("GET", url) as resp:
+                with dst.open("wb") as f:
+                    f.write(await resp.read())
+                    return True
+        except aiohttp.ClientError:
+            return False
+
+    # Protocol implementation
     def gen_msg_id(self) -> int:
         self._msg_id += 1
         return self._msg_id
 
     async def init_adoption(self) -> None:
         self.logger.info(
-            f"Initiating adoption with token [{self.args.token}] and mac [{self.args.mac}]"
+            f"Adopting with token [{self.args.token}] and mac [{self.args.mac}]"
         )
         await self.send(
             self.gen_response(
@@ -179,7 +221,7 @@ class UnifiCamBase(metaclass=ABCMeta):
                     "totalLoad": 0.5474,
                     "upgradeTimeoutSec": 150,
                     "uptime": self.get_uptime(),
-                    "features": {},
+                    "features": self.get_feature_flags(),
                 },
             ),
         )
@@ -188,7 +230,10 @@ class UnifiCamBase(metaclass=ABCMeta):
         return self.gen_response(
             "ubnt_avclient_paramAgreement",
             msg["messageId"],
-            {"authToken": self.args.token, "features": {}},
+            {
+                "authToken": self.args.token,
+                "features": self.get_feature_flags(),
+            },
         )
 
     async def process_upgrade(self, msg: AVClientRequest) -> None:
@@ -687,8 +732,12 @@ class UnifiCamBase(metaclass=ABCMeta):
     async def process_snapshot_request(
         self, msg: AVClientRequest
     ) -> Optional[AVClientResponse]:
+        snapshot_type = msg["payload"]["what"]
+        if snapshot_type in ["motionSnapshot", "smartDetectZoneSnapshot"]:
+            path = self._motion_snapshot
+        else:
+            path = await self.get_snapshot()
 
-        path = await self.get_snapshot()
         if os.path.isfile(path):
             async with aiohttp.ClientSession() as session:
                 files = {"payload": open(path, "rb")}
@@ -699,6 +748,7 @@ class UnifiCamBase(metaclass=ABCMeta):
                         data=files,
                         ssl=self._ssl_context,
                     )
+                    self.logger.debug(f"Uploaded {snapshot_type} from {path}")
                 except aiohttp.ClientError:
                     self.logger.exception("Failed to upload snapshot")
         else:
@@ -721,8 +771,10 @@ class UnifiCamBase(metaclass=ABCMeta):
         )
 
     def gen_response(
-        self, name: str, response_to: int = 0, payload: Dict[str, Any] = {}
+        self, name: str, response_to: int = 0, payload: Optional[Dict[str, Any]] = None
     ) -> AVClientResponse:
+        if not payload:
+            payload = {}
         return {
             "from": "ubnt_avclient",
             "functionName": name,
@@ -751,7 +803,7 @@ class UnifiCamBase(metaclass=ABCMeta):
 
         if (
             ("responseExpected" not in m)
-            or (m["responseExpected"] == False)
+            or (m["responseExpected"] is False)
             and (
                 fn
                 not in [
@@ -794,6 +846,10 @@ class UnifiCamBase(metaclass=ABCMeta):
             res = self.gen_response(
                 "UpdateUsernamePassword", response_to=m["messageId"]
             )
+        elif fn == "ChangeSmartDetectSettings":
+            res = self.gen_response(
+                "ChangeSmartDetectSettings", response_to=m["messageId"]
+            )
         elif fn == "UpdateFirmwareRequest":
             await self.process_upgrade(m)
             return True
@@ -813,7 +869,14 @@ class UnifiCamBase(metaclass=ABCMeta):
             or self._ffmpeg_handles[stream_index].poll() is not None
         ):
             source = self.get_stream_source(stream_index)
-            cmd = f'ffmpeg -nostdin -y -rtsp_transport {self.args.rtsp_transport} -i "{source}" {self.args.ffmpeg_args} -metadata streamname={stream_name} -f flv - | {sys.executable} -m unifi.clock_sync | nc {destination[0]} {destination[1]}'
+            cmd = (
+                "ffmpeg -nostdin -y -stimeout 15000000"
+                f" -rtsp_transport {self.args.rtsp_transport}"
+                f' -i "{source}" {self.args.ffmpeg_args}'
+                f" -metadata streamname={stream_name} -f flv -"
+                f" | {sys.executable} -m unifi.clock_sync"
+                f" | nc {destination[0]} {destination[1]}"
+            )
             self.logger.info(
                 f"Spawning ffmpeg for {stream_index} ({stream_name}): {cmd}"
             )
