@@ -6,18 +6,22 @@ from shutil import which
 
 import coloredlogs
 
-from unifi.cams import HikvisionCam, LorexCam, RTSPCam
+from unifi.cams import FrigateCam, HikvisionCam, LorexCam, ReolinkNVRCam, RTSPCam
 from unifi.core import Core
+from unifi.version import __version__
 
 CAMS = {
+    "frigate": FrigateCam,
     "hikvision": HikvisionCam,
     "lorex": LorexCam,
+    "reolink_nvr": ReolinkNVRCam,
     "rtsp": RTSPCam,
 }
 
 
 def parse_args():
     parser = argparse.ArgumentParser()
+    parser.add_argument("--version", action="version", version=__version__)
     parser.add_argument("--host", "-H", required=True, help="NVR ip address and port")
     parser.add_argument(
         "--cert",
@@ -91,7 +95,7 @@ def main():
 
     # Preflight checks
     if which("ffmpeg") is None:
-        self.logger.error("ffmpeg is not installed")
+        logger.error("ffmpeg is not installed")
         sys.exit(1)
 
     cam = klass(args, logger)
